#!/usr/bin/python

DOCUMENTATION = r'''
---
module: decort_lb

description: See L(Module Documentation,https://repository.basistech.ru/BASIS/decort-ansible/wiki/Home).
'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.basic import env_fallback
from ansible.module_utils.decort_utils import *


class decort_lb(DecortController):
    def __init__(self) -> None:
        super(decort_lb,self).__init__(AnsibleModule(**self.amodule_init_args))
        arg_amodule = self.amodule

        self.lb_id = 0
        self.lb_facts = None
        self.vins_id = 0
        self.vins_facts = None
        self.rg_id = 0
        self.rg_facts = None
        self.default_server_check = "enabled"
        self.default_alg = "roundrobin"
        self.default_settings = {
                "downinter": 10000,
                "fall": 2,
                "inter": 5000,
                "maxconn": 250,
                "maxqueue": 256,
                "rise": 2,
                "slowstart": 60000,
                "weight": 100,
        }
        self.is_lb_stopped_or_will_be_stopped: None | bool = None

        if arg_amodule.params['lb_id']:
            self.lb_id, self.lb_facts = self.lb_find(arg_amodule.params['lb_id'])
            if not self.lb_id:
                self.result['failed'] = True
                self.result['msg'] = "Specified LB ID {} not found."\
                .format(arg_amodule.params['lb _id'])
                self.fail_json(**self.result)
            self.rg_id = self.lb_facts['rgId']
            self.vins_id = self.lb_facts['vinsId']

        elif arg_amodule.params['rg_id']:
            self.rg_id, self.rg_facts = self.rg_find(0,arg_amodule.params['rg_id'], arg_rg_name="")
            if not self.rg_id:
                self.result['failed'] = True
                self.result['msg'] = "Specified RG ID {} not found.".format(arg_amodule.params['vins_id'])
                self.amodule.fail_json(**self.result)
            self.acc_id = self.rg_facts['accountId']

        elif arg_amodule.params['account_id'] or arg_amodule.params['account_name'] != "":
            
            if not arg_amodule.params['rg_name']:
                self.result['failed'] = True
                self.result['msg'] = ("RG name must be specified with account present")
                self.amodule.fail_json(**self.result)
            self.acc_id, self._acc_info = self.account_find(arg_amodule.params['account_name'], 
                                                            arg_amodule.params['account_id'])
            if not self.acc_id:
                self.result['failed'] = True
                self.result['msg'] = ("Current user does not have access to the requested account "
                                    "or non-existent account specified.")
                self.amodule.fail_json(**self.result)
            self.rg_id, self.rg_facts = self.rg_find(self.acc_id,0, arg_rg_name=arg_amodule.params['rg_name'])
        
        if arg_amodule.params['vins_id']:
            self.vins_id, self.vins_facts = self.vins_find(
                vins_id=arg_amodule.params['vins_id']
                )
            if not self.vins_id:
                self.result['failed'] = True
                self.result['msg'] = (
                    f'Specified ViNS ID {arg_amodule.params["vins_id"]}'
                    f' not found'
                    )
                self.amodule.fail_json(**self.result)
        elif arg_amodule.params['vins_name']:
            self.vins_id, self.vins_facts = self.vins_find(
                vins_id=arg_amodule.params['vins_id'],
                vins_name=arg_amodule.params['vins_name'],
                rg_id=self.rg_id)
            if not self.vins_id:
                self.result['failed'] = True
                self.result['msg'] = (
                    f'Specified ViNS name {arg_amodule.params["vins_name"]}'
                    f' not found in RG ID {self.rg_id}'
                    )
                self.amodule.fail_json(**self.result)

        if self.rg_id and arg_amodule.params['lb_name']:
            self.lb_id, self.lb_facts = self.lb_find(0,arg_amodule.params['lb_name'],self.rg_id)

        if self.lb_id and self.lb_facts['status'] != 'DESTROYED':
            self.acc_id = self.lb_facts['accountId']
            self.check_amodule_args_for_change()
        else:
            self.check_amodule_args_for_create()

        return

    def create(self):
        start_after_create = self.aparams['state'] != 'stopped'
        self.lb_id = self.lb_provision(self.amodule.params['lb_name'],
                                    self.rg_id,self.vins_id,
                                    self.amodule.params['ext_net_id'],
                                    self.amodule.params['ha_lb'],
                                    self.amodule.params['description'],
                                    sysctl=self.amodule.params['sysctl'],
                                    zone_id=self.aparams['zone_id'],
                                    start=start_after_create,
        )
        if self.lb_id and (self.amodule.params['backends'] or
                           self.amodule.params['frontends']):
            self.lb_id, self.lb_facts = self.lb_find(0,self.amodule.params['lb_name'],self.rg_id)
            self.lb_update(
                lb_facts=self.lb_facts,
                aparam_backends=self.amodule.params['backends'],
                aparam_frontends=self.amodule.params['frontends'],
                aparam_servers=self.amodule.params['servers'],
            )
        return
    
    def action(self,d_state='',restore=False):
        if restore == True:
            self.lb_restore(lb_id=self.lb_id)
            _, self.lb_facts = self._lb_get_by_id(lb_id=self.lb_id)
            self.lb_state(self.lb_facts, 'enabled')
            _, self.lb_facts = self._lb_get_by_id(lb_id=self.lb_id)
        
        self.lb_update(
            lb_facts=self.lb_facts,
            aparam_backends=self.amodule.params['backends'],
            aparam_frontends=self.amodule.params['frontends'],
            aparam_servers=self.amodule.params['servers'],
            aparam_sysctl=self.aparams['sysctl'],
        )

        if d_state != '':
            self.lb_state(self.lb_facts, d_state)
            _, self.lb_facts = self._lb_get_by_id(lb_id=self.lb_id)

        if (d_state == 'enabled' and
                self.lb_facts.get('status') == 'ENABLED' and
                self.lb_facts.get('techStatus') == 'STOPPED'):
            self.lb_state(self.lb_facts, 'started')
            _, self.lb_facts = self._lb_get_by_id(lb_id=self.lb_id)

        aparam_zone_id = self.aparams['zone_id']
        if aparam_zone_id is not None and aparam_zone_id != self.lb_facts['zoneId']:
            self.lb_migrate_to_zone(
                lb_id=self.lb_id,
                zone_id=aparam_zone_id,
            )

        return

    def delete(self):
        self.lb_delete(self.lb_id, self.amodule.params['permanently'])
        self.lb_facts['status'] = 'DESTROYED'
        return
    def nop(self):
        """No operation (NOP) handler for LB management by decort_lb module.
        This function is intended to be called from the main switch construct of the module
        when current state -> desired state change logic does not require any changes to
        the actual LB state.
        """
        self.result['failed'] = False
        self.result['changed'] = False
        if self.lb_id:
            self.result['msg'] = ("No state change required for LB  ID {} because of its "
                                   "current status '{}'.").format(self.lb_id, self.lb_facts['status'])
        else:
            self.result['msg'] = ("No state change to '{}' can be done for "
                                  "non-existent LB instance.").format(self.amodule.params['state'])
        return
    def error(self):
        self.result['failed'] = True
        self.result['changed'] = False
        if self.vins_id:
            self.result['failed'] = True
            self.result['changed'] = False
            self.result['msg'] = ("Invalid target state '{}' requested for LB ID {} in the "
                                    "current status '{}'").format(self.lb_id,
                                                                    self.amodule.params['state'],
                                                                    self.lb_facts['status'])
        else:
            self.result['failed'] = True
            self.result['changed'] = False
            self.result['msg'] = ("Invalid target state '{}' requested for non-existent "
                                   "LB name '{}'").format(self.amodule.params['state'],
                                                            self.amodule.params['lb_name'])   
        return
    def package_facts(self, arg_check_mode=False):
        """Package a dictionary of LB facts according to the decort_lb module specification. 
        This dictionary will be returned to the upstream Ansible engine at the completion of 
        the module run.

        @param arg_check_mode: boolean that tells if this Ansible module is run in check mode
        """

        ret_dict = dict(id=0,
                        name="none",
                        state="CHECK_MODE",
                        sysctl={},
                        )

        if arg_check_mode:
            # in check mode return immediately with the default values
            return ret_dict

        if self.lb_facts is None:
            # if void facts provided - change state value to ABSENT and return
            ret_dict['state'] = "ABSENT"
            return ret_dict

        ret_dict['id'] = self.lb_facts['id']
        ret_dict['name'] = self.lb_facts['name']
        ret_dict['state'] = self.lb_facts['status']
        ret_dict['account_id'] = self.lb_facts['accountId']
        ret_dict['rg_id'] = self.lb_facts['rgId']
        ret_dict['gid'] = self.lb_facts['gid']
        if self.amodule.params['state']!="absent":
            ret_dict['backends'] = self.lb_facts['backends']
            ret_dict['frontends'] = self.lb_facts['frontends']
        ret_dict['sysctl'] = self.lb_facts['sysctlParams']
        ret_dict['zone_id'] = self.lb_facts['zoneId']
        ret_dict['tech_status'] = self.lb_facts['techStatus']
        return ret_dict

    @property
    def amodule_init_args(self) -> dict:
        return self.pack_amodule_init_args(
            argument_spec=dict(
                account_id=dict(
                    type='int',
                ),
                account_name=dict(
                    type='str',
                    default='',
                ),
                description=dict(
                    type='str',
                ),
                ext_net_id=dict(
                    type='int',
                    default=0,
                ),
                ext_ip_addr=dict(
                    type='str',
                    default='',
                ),
                state=dict(
                    type='str',
                    choices=[
                        'absent',
                        'disabled',
                        'enabled',
                        'present',
                        'restart',
                        'started',
                        'stopped',
                    ],
                ),
                rg_id=dict(
                    type='int',
                    default=0,
                ),
                rg_name=dict(
                    type='str',
                    default='',
                ),
                vins_name=dict(
                    type='str',
                    default='',
                ),
                vins_id=dict(
                    type='int',
                    default=0,
                ),
                lb_id=dict(
                    type='int',
                    default=0,
                ),
                lb_name=dict(
                    type='str',
                ),
                ha_lb=dict(
                    type='bool',
                    default=False,
                ),
                backends=dict(
                    type='list',
                ),
                frontends=dict(
                    type='list',
                ),
                servers=dict(
                    type='list',
                ),
                permanently=dict(
                    type='bool',
                    default=False,
                ),
                sysctl=dict(
                    type='dict',
                ),
                zone_id=dict(
                    type=int,
                ),
            ),
            supports_check_mode=True,
            required_one_of=[
                ('rg_id', 'rg_name'),
                ('lb_id', 'lb_name'),
                ('vins_id', 'vins_name'),
            ],
        )

    def check_amodule_args_for_change(self):
        check_errors = False

        lb_info: dict = self.lb_facts
        self.is_lb_stopped_or_will_be_stopped  = (
            (
                lb_info['techStatus'] == 'STOPPED'
                and (
                    self.aparams['state'] is None
                    or self.aparams['state'] in ('present', 'stopped')
                )
            )
            or (
                lb_info['techStatus'] != 'STOPPED'
                and self.aparams['state'] == 'stopped'
            )
        )

        if self.check_aparam_zone_id() is False:
            check_errors = True
        if (
            self.aparams['zone_id'] is not None
            and self.aparams['zone_id'] != lb_info['zoneId']
            and not self.is_lb_stopped_or_will_be_stopped
        ):
            check_errors = True
            self.message(
                'Check for parameter "zone_id" failed: '
                'Load balancer must be stopped to migrate to a zone.'
            )

        if check_errors:
            self.exit(fail=True)

    def check_amodule_args_for_create(self):
        check_errors = False
        if self.check_aparam_zone_id() is False:
            check_errors = True

        if check_errors:
            self.exit(fail=True)


def main():
    decon = decort_lb()
    amodule = decon.amodule
    if decon.lb_id:
        if decon.lb_facts['status'] in ["MODELED", "DISABLING", "ENABLING", "DELETING","DESTROYING","RESTORING"]:
            decon.result['failed'] = True
            decon.result['changed'] = False
            decon.result['msg'] = ("No change can be done for existing LB ID {} because of its current "
                                "status '{}'").format(decon.lb_id, decon.lb_facts['status'])
        elif decon.lb_facts['status'] in ('DISABLED', 'ENABLED', 'CREATED'):
            if amodule.params['state'] == 'absent':
                decon.delete()
            else:
                decon.action(d_state=amodule.params['state'])
        elif decon.lb_facts['status'] == "DELETED":
            if amodule.params['state'] == 'present':
                decon.action(restore=True)
            elif amodule.params['state'] == 'enabled':
                decon.action(d_state='enabled', restore=True)
            elif (amodule.params['state'] == 'absent' and
                  amodule.params['permanently']):
                decon.delete()
            elif amodule.params['state'] == 'disabled':
                decon.error()
        elif decon.lb_facts['status'] == "DESTROYED":
            if amodule.params['state'] in ('present', 'enabled'):
                decon.create()
            elif amodule.params['state'] == 'absent':
                decon.nop()
            elif amodule.params['state'] == 'disabled':
                decon.error()
    else:
        state = amodule.params['state']
        if state is None:
            state = 'present'
        if state == 'absent':
            decon.nop()
        elif state in ('present', 'enabled', 'stopped', 'started'):
            decon.create()
        elif state == 'disabled':
            decon.error()

    if decon.result['failed']:
        amodule.fail_json(**decon.result)
    else:
        if decon.result['changed'] and amodule.params['state'] != 'absent':
            _, decon.lb_facts = decon.lb_find(decon.lb_id)
        if decon.lb_id:
            decon.result['facts'] = decon.package_facts(amodule.check_mode)
        amodule.exit_json(**decon.result)
if __name__ == "__main__":
    main()
